pub struct API {
    inner: connection_capnp::bootstrap::Client,
    log: Logger,
}

impl Api {
    fn new(log: Logger, inner: connection_capnp::bootstrap::Client) -> Self {
        Self { inner}
    }
    pub fn from_stream(stream: TcpStream) -> (impl Future, Self) {
        let network = Box::new(twoparty::VatNetwork::new(stream.clone(), stream.clone(),
                rpc_twoparty_capnp::Side::Client, Default::default()));

        let mut rpc_system = RpcSystem::new(network, None); 
        let bffh: connection_capnp::bootstrap::Client 
            = rpc_system.bootstrap(rpc_twoparty_capnp::Side::Server);

        (rpc_system, Api::new(bffh))
    }

    pub async fn authentication(&mut self) -> Authentication {
        let req = self.bffh.auth_request().send().promise;
        // TODO: When's that an Err?
        let res = req.await.unwrap();
        // TODO: When's that an Err?
        let tmp = res.get().unwrap();
        Authentication::new(tmp.get_auth().unwrap())
    }

    async fn authenticate(&mut self, log: Logger) {
        let r = self.bffh.auth_request();
        let auth = r.send().pipeline.get_auth();
        let m = auth.mechanisms_request().send().promise.await.unwrap();

        for t in m.get().unwrap().get_mechs().unwrap().iter() {
            info!(log, "Mechanism {} available", t.unwrap());
        }
    }

    ///// Authenticate to the server. Returns true on success, false on error
    //async fn authenticate(&mut self) -> Result<bool, io::Error> {
    //    let mut sasl = SASL::new().unwrap();
    //    let plain = std::ffi::CString::new("PLAIN").unwrap();
    //    let mut sess = sasl.client_start(&plain).unwrap();
    //    sess.set_property(rsasl::Property::GSASL_AUTHID, b"testuser");
    //    sess.set_property(rsasl::Property::GSASL_PASSWORD, b"testpass");

    //    if let rsasl::Step::Done(data) = sess.step(&[]).unwrap() {
    //        self.send_authentication_request("PLAIN", Some(&data)).await;
    //    } else {
    //        println!("Sasl said moar data");
    //    }

    //    Ok(self.receive_challenge().await?)
    //}

    //fn send_authentication_request(&mut self, mech: &str, init: Option<&[u8]>) -> impl Future<Output=()> {
    //    let mut outer = capnp::message::Builder::new_default();
    //    let mut builder = outer.init_root::<connection_capnp::message::Builder>()
    //        .init_auth()
    //        .init_request();
    //    builder.set_mechanism(mech);

    //    if let Some(data) = init {
    //        builder.init_initial_response().set_initial(data);
    //    }

    //    let stream = self.stream.clone();
    //    capnp_futures::serialize::write_message(stream, outer).map(|r| r.unwrap())
    //}

    //async fn receive_challenge(&mut self) -> Result<bool, io::Error> {
    //    let message = capnp_futures::serialize::read_message(&mut self.stream, capnp::message::ReaderOptions::default()).await.unwrap().unwrap();
    //    let m = message.get_root::<connection_capnp::message::Reader>().unwrap();

    //    if let Ok(connection_capnp::message::Which::Auth(Ok(r))) = m.which() {
    //        if let Ok(auth_capnp::auth_message::Outcome(Ok(r))) = r.which() {
    //            if let Ok(auth_capnp::outcome::Result::Successful) = r.get_result() {
    //                return Ok(true);
    //            }
    //        }
    //    }

    //    return Ok(false);
    //}
}
